import uuid
from datetime import datetime

from spce import CloudEvent


class EventSource:

    def __init__(self, client, *,
                 type="",
                 source="",
                 datacontenttype="application/json",
                 dataschema="",
                 subject=""
         ):
        self.client = client
        self.type = type
        self.source = source or str(uuid.uuid4())
        self.datacontenttype = datacontenttype
        self.dataschema = dataschema
        self.subject=subject
        self._id_seq = 1

    def post_event(self, *,
              type= "",
              source= "",
              id="",
              subject="",
              time="",
              datacontenttype="application/json",
              dataschema="",
              data="") -> CloudEvent:
        type = type or self.type
        source = source or self.source
        id = str(id) if id else str(self._id_seq)
        time = time or datetime.utcnow()
        datacontenttype=datacontenttype or self.datacontenttype

        event = CloudEvent(
            type=type,
            source=source,
            id=id,
            time=time,
            subject=subject,
            data=data,
            datacontenttype=datacontenttype if data else "",
            dataschema=dataschema if data else "",
        )

        # TODO: make this thread-safe
        self._id_seq += 1
        self.client.post_event(event)

        return event






import json
from typing import List

import urllib3

from .http import Http
from .mqtt import Mqtt


class Client:

    pool = urllib3.PoolManager(
        retries=urllib3.Retry(connect=3, read=2, redirect=2, status=0))

    def __init__(self, *,
                 org_id: str,
                 app_id: str,
                 api_key: str,
                 beacon_host="beacon.scaleplan.net",
                 beacon_port=9000):
        self.org_id = org_id
        self.app_id = app_id
        self.api_key = api_key
        self.beacon_addr = beacon_host
        self.http_url = ""
        self.auth_header = f"Bearer {beacon_host}"
        self.beacon_url = f"http://{beacon_host}:{beacon_port}/v0/info"
        self._server = None
        self._http = None
        self._mqtt = None
        self.connect_beacon()


    @property
    def http(self) -> Http:
        return self._http

    @property
    def mqtt(self) -> Mqtt:
        return self._mqtt

    def connect_beacon(self) -> (dict, List):
        body = json.dumps({
            "org_id": self.org_id,
            "app_id": self.app_id,
        })
        headers = {
            "Authorization": self.auth_header,
            "Content-Type": "application/json",
        }
        r = self.pool.request("POST", self.beacon_url, headers=headers, body=body)
        text = r.data.decode("utf-8")
        if r.status != 200:
            raise ClientError("Could not receive beacon info %s" % text)
        resp_json = json.loads(text)
        self._server = resp_json["server"]
        self.process_features(resp_json["features"])

    def process_features(self, features):
        http_feature = features.get("http")
        if http_feature is not None and len(http_feature) > 0:
            # TODO: check whether JSON is supported
            http_feature = http_feature[0]
            proto = "https" if http_feature.get("ssl") else "http"
            host = http_feature.get("host")
            if host is None:
                raise ClientError("Invalid HTTP feature: host is missing")
            port = http_feature.get("port")
            if port is None:
                raise ClientError("Invalid HTTP feature: port is missing")
            base_url = f"{proto}://{host}:{port}"
            self._http = Http(self.pool, base_url)
            return

        mqtt_feature = features.get("mqtt")
        if mqtt_feature is not None and len(mqtt_feature) > 0:
            # TODO: check whether JSON is supported
            mqtt_feature = mqtt_feature[0]
            host = mqtt_feature.get("host")
            if host is None:
                raise ClientError("Invalid MQTT feature: host is missing")
            port = mqtt_feature.get("port")
            if port is None:
                raise ClientError("Invalid MQTT feature: port is missing")
            self._mqtt = Mqtt(host=host, port=port)
            return

        raise ClientError("HTTP and MQTT are not supported by the gateway.")


class ClientError(Exception):
    pass


from spce import CloudEvent, Json

from .source import EventSource


class Http:

    def __init__(self, client, base_url, path = ""):
        self.client = client
        self.base_url = base_url
        self.path = path or "/event"

    def event_source(self, path: str="", *,
                     type="",
                     source="",
                     datacontenttype="application/json",
                     dataschema="",
                     subject="") -> EventSource:
        client = Http(self.client, self.base_url, f'/{path.strip("/")}')
        return EventSource(client,
                           type=type,
                           source=source,
                           datacontenttype=datacontenttype,
                           dataschema=dataschema,
                           subject=subject)

    def post_event(self, event: CloudEvent, path = ""):
        path = path or self.path
        body, headers = self._make_body_headers(event, "json")
        url = f"{self.base_url}{path}"
        return self.client.request("POST", url, headers=headers, body=body)

    @classmethod
    def _make_body_headers(self, event, format):
        if format == "json":
            headers = {
                "content-type": "application/cloudevents+json"
            }
            body = Json.encode(event)
        elif format == "avro":
            headers = {
                "content-type": "application/cloudevents+avro"
            }
            body = Avro.encode(event)
        else:
            raise RuntimeError("_make_headers: %s is not a valid format" % format)

        return body, headers



import urllib.parse

from .client import Client


class Node:

    def __init__(self, name: str, class_: str, instances: int, config: dict):
        self.name = name
        self.class_ = class_
        self.instances = instances
        self.config = config
        self._next_node = None

    def __rshift__(self, other: "Node") -> "Node":
        print(f"{self.name} >> {other.name}")
        self._next_node = other
        self.config["forward"] = other.name
        return other

    def _as_config(self):
        config = {
            "instances": self.instances
        }
        if self.class_:
            config["class"] = self.class_
        config["config"] = self.config
        return config


class Gateway:

    def __init__(self, addr="localhost:9000", *, org_id="default", app_id="default", api_key: str):
        parsed_addr = urllib.parse.urlparse(addr)
        self.org_id = org_id
        self.app_id = app_id
        self.api_key = api_key
        self.beacon_host = parsed_addr.hostname
        self.beacon_port = parsed_addr.port
        self.start_nodes = []

    @classmethod
    def _beacon_addr(cls, addr: str) -> (str, int):
        for proto in ["https://", "http://"]:
            if addr.startswith(proto):
                addr = addr[len(proto):]
                break
        if addr.find("://") >= 0:
            raise RuntimeError(f"Invalid beacon address: {addr}")
        try:
            if ":" in addr:
                host, port = addr.split(":", 2)
                host = host or "localhost"
                port = int(port)
                return host, port
            host = addr
            return host, 9000
        except (ValueError, TypeError):
            raise RuntimeError(f"Invalid beacon address: {addr}")

    def node(self, name, class_="", instances=1, **config) -> Node:
        node = Node(name, class_, instances, config)
        self.start_nodes.append(node)
        return node

    def update(self):
        pass

    def client(self) -> Client:
        return Client(
            org_id=self.org_id,
            app_id=self.app_id,
            api_key=self.api_key,
            beacon_host=self.beacon_host,
            beacon_port=self.beacon_port
        )

    def _config(self):
        nodes = {}
        for start_node in self.start_nodes:
            node = start_node
            while node:
                nodes[node.name] = node._as_config()
                node = node._next_node
        return {
            "nodes": nodes,
        }

    def __enter__(self):
        print("CTX ENTER")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        print("CTX EXIT")
        print(f"config: {self._config()}")
        # self.update()



# Copyright 2020 Scale Plan Yazılım A.Ş.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name='pdgw',
    version='0.2.0',
    packages=['pdgw'],
    url='https://scaleplan.io',
    license='Proprietary',
    author='ScalePlan Engineering',
    author_email='dev@scaleplan.io',
    description="Particle DataGateway client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
    ],
    keywords='particle,datagateway,cloudevents,scaleplan,cloud,events,event',
    install_requires=['spce', 'urllib3'],
    tests_require=['pytest', 'coverage', 'pytest-cov', 'avro', 'paho-mqtt'],
    extras_require={
        'avro': ['spce[avro]>=0.2.0'],
        'mqtt': ['paho-mqtt>=1.5.1']
    }
)




import paho.mqtt.client as mqtt
from spce import CloudEvent, Json

from .source import EventSource


class Mqtt:

    def __init__(self, host: str, port: int, topic = ""):
        self.host = host
        self.port = port
        self.topic = topic
        self.mqtt_client = mqtt.Client()
        # client.on_connect
        # client.on_message
        self.mqtt_client.connect(host=host, port=port)

    def event_source(self, topic: str, *,
                     type="",
                     source="",
                     datacontenttype="application/json",
                     dataschema="",
                     subject="") -> EventSource:
        client = Mqtt(self.host, self.port, topic)
        return EventSource(client,
                           type=type,
                           source=source,
                           datacontenttype=datacontenttype,
                           dataschema=dataschema,
                           subject=subject)

    def post_event(self, event: CloudEvent, topic = "", *, qos=0):
        topic = topic or self.topic
        # TODO: content type from /v0/info
        payload = Json.encode(event)
        return self.mqtt_client.publish(topic, payload, qos=qos)
